from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from .base import SQLObject
from .expression import Expression
from .enum import Enum
from .note import Note
from pydbml.exceptions import TableNotFoundError
from pydbml.tools import comment_to_dbml
from pydbml.tools import comment_to_sql
from pydbml.tools import note_option_to_dbml

if TYPE_CHECKING:  # pragma: no cover
    from .table import Table
    from .reference import Reference


class Column(SQLObject):
    '''Class representing table column.'''

    required_attributes = ('name', 'type')

    def __init__(self,
                 name: str,
                 type: Union[str, Enum],
                 unique: bool = False,
                 not_null: bool = False,
                 pk: bool = False,
                 autoinc: bool = False,
                 default: Optional[Union[str, int, bool, float, Expression]] = None,
                 note: Optional[Union['Note', str]] = None,
                 # ref_blueprints: Optional[List[ReferenceBlueprint]] = None,
                 comment: Optional[str] = None):
        self.name = name
        self.type = type
        self.unique = unique
        self.not_null = not_null
        self.pk = pk
        self.autoinc = autoinc
        self.comment = comment
        self.note = Note(note)

        self.default = default
        self.table: Optional['Table'] = None

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    def get_refs(self) -> List['Reference']:
        '''
        get all references related to this column (where this col is col1 in)
        '''
        if not self.table:
            raise TableNotFoundError('Table for the column is not set')
        return [ref for ref in self.table.get_refs() if self in ref.col1]

    @property
    def database(self):
        return self.table.database if self.table else None

    @property
    def sql(self):
        '''
        Returns inline SQL of the column, which should be a part of table definition:

        "id" integer PRIMARY KEY AUTOINCREMENT
        '''

        self.check_attributes_for_sql()
        components = [f'"{self.name}"']
        if isinstance(self.type, Enum):
            components.append(self.type._get_full_name_for_sql())
        else:
            components.append(str(self.type))

        table_has_composite_pk = False if self.table is None else self.table._has_composite_pk()
        if self.pk and not table_has_composite_pk:  # comp-PKs are rendered in table sql
            components.append('PRIMARY KEY')
        if self.autoinc:
            components.append('AUTOINCREMENT')
        if self.unique:
            components.append('UNIQUE')
        if self.not_null:
            components.append('NOT NULL')
        if self.default is not None:
            default = self.default.sql \
                if isinstance(self.default, Expression) else self.default
            components.append(f'DEFAULT {default}')

        result = comment_to_sql(self.comment) if self.comment else ''
        result += ' '.join(components)
        return result

    @property
    def dbml(self):
        def default_to_str(val: Union[Expression, str]) -> str:
            if isinstance(val, str):
                if val.lower() in ('null', 'true', 'false'):
                    return val.lower()
                else:
                    return f"'{val}'"
            elif isinstance(val, Expression):
                return val.dbml
            else:  # int or float or bool
                return val

        result = comment_to_dbml(self.comment) if self.comment else ''
        result += f'"{self.name}" '
        if isinstance(self.type, Enum):
            result += self.type._get_full_name_for_sql()
        else:
            result += self.type

        options = [ref.dbml for ref in self.get_refs() if ref.inline]
        if self.pk:
            options.append('pk')
        if self.autoinc:
            options.append('increment')
        if self.default:
            options.append(f'default: {default_to_str(self.default)}')
        if self.unique:
            options.append('unique')
        if self.not_null:
            options.append('not null')
        if self.note:
            options.append(note_option_to_dbml(self.note))

        if options:
            result += f' [{", ".join(options)}]'
        return result

    def __repr__(self):
        '''
        >>> Column('name', 'VARCHAR2')
        <Column 'name', 'VARCHAR2'>
        '''
        type_name = self.type if isinstance(self.type, str) else self.type.name
        return f'<Column {self.name!r}, {type_name!r}>'

    def __str__(self):
        '''
        >>> print(Column('name', 'VARCHAR2'))
        name[VARCHAR2]
        '''

        return f'{self.name}[{self.type}]'



from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union
from typing import Iterable

from .base import SQLObject
from .column import Column
from .index import Index
from .note import Note
from pydbml.constants import MANY_TO_ONE
from pydbml.constants import ONE_TO_MANY
from pydbml.constants import ONE_TO_ONE
from pydbml.exceptions import ColumnNotFoundError
from pydbml.exceptions import IndexNotFoundError
from pydbml.exceptions import UnknownDatabaseError
from pydbml.tools import comment_to_dbml
from pydbml.tools import comment_to_sql
from pydbml.tools import indent


if TYPE_CHECKING:  # pragma: no cover
    from pydbml.database import Database
    from .reference import Reference


class Table(SQLObject):
    '''Class representing table.'''

    required_attributes = ('name', 'schema')

    def __init__(self,
                 name: str,
                 schema: str = 'public',
                 alias: Optional[str] = None,
                 columns: Optional[Iterable[Column]] = None,
                 indexes: Optional[Iterable[Index]] = None,
                 note: Optional[Union['Note', str]] = None,
                 header_color: Optional[str] = None,
                 comment: Optional[str] = None,
                 abstract: bool = False):
        self.database: Optional[Database] = None
        self.name = name
        self.schema = schema
        self.columns: List[Column] = []
        for column in columns or []:
            self.add_column(column)
        self.indexes: List[Index] = []
        for index in indexes or []:
            self.add_index(index)
        self.alias = alias if alias else None
        self.note = Note(note)
        self.header_color = header_color
        self.comment = comment
        self.abstract = abstract

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    @property
    def full_name(self) -> str:
        return f'{self.schema}.{self.name}'

    def _has_composite_pk(self) -> bool:
        return sum(c.pk for c in self.columns) > 1

    def add_column(self, c: Column) -> None:
        '''
        Adds column to self.columns attribute and sets in this column the
        `table` attribute.
        '''
        if not isinstance(c, Column):
            raise TypeError('Columns must be of type Column')
        c.table = self
        self.columns.append(c)

    def delete_column(self, c: Union[Column, int]) -> Column:
        if isinstance(c, Column):
            if c in self.columns:
                c.table = None
                return self.columns.pop(self.columns.index(c))
            else:
                raise ColumnNotFoundError(f'Column {c} if missing in the table')
        elif isinstance(c, int):
            self.columns[c].table = None
            return self.columns.pop(c)

    def add_index(self, i: Index) -> None:
        '''
        Adds index to self.indexes attribute and sets in this index the
        `table` attribute.
        '''
        if not isinstance(i, Index):
            raise TypeError('Indexes must be of type Index')
        for subject in i.subjects:
            if isinstance(subject, Column) and subject.table != self:
                raise ColumnNotFoundError(f'Column {subject} not in the table')
        i.table = self
        self.indexes.append(i)

    def delete_index(self, i: Union[Index, int]) -> Index:
        if isinstance(i, Index):
            if i in self.indexes:
                i.table = None
                return self.indexes.pop(self.indexes.index(i))
            else:
                raise IndexNotFoundError(f'Index {i} if missing in the table')
        elif isinstance(i, int):
            self.indexes[i].table = None
            return self.indexes.pop(i)

    def get_refs(self) -> List['Reference']:
        if not self.database:
            raise UnknownDatabaseError('Database for the table is not set')
        return [ref for ref in self.database.refs if ref.table1 == self]

    def _get_references_for_sql(self) -> List['Reference']:
        '''
        return inline references for this table sql definition
        '''
        if self.abstract:
            return []
        if not self.database:
            raise UnknownDatabaseError(f'Database for the table {self} is not set')
        result = []
        for ref in self.database.refs:
            if ref.inline:
                if (ref.type in (MANY_TO_ONE, ONE_TO_ONE)) and\
                        (ref.table1 == self):
                    result.append(ref)
                elif (ref.type == ONE_TO_MANY) and (ref.table2 == self):
                    result.append(ref)
        return result

    def _get_full_name_for_sql(self) -> str:
        if self.schema == 'public':
            return f'"{self.name}"'
        else:
            return f'"{self.schema}"."{self.name}"'

    def __getitem__(self, k: Union[int, str]) -> Column:
        if isinstance(k, int):
            return self.columns[k]
        elif isinstance(k, str):
            for c in self.columns:
                if c.name == k:
                    return c
            raise ColumnNotFoundError(f'Column {k} not present in table {self.name}')
        else:
            raise TypeError('indeces must be str or int')

    def get(self, k, default: Optional[Column] = None) -> Optional[Column]:
        try:
            return self.__getitem__(k)
        except (IndexError, ColumnNotFoundError):
            return default

    def __iter__(self):
        return iter(self.columns)

    def __repr__(self):
        '''
        >>> table = Table('customers')
        >>> table
        <Table 'public' 'customers'>
        '''

        return f'<Table {self.schema!r} {self.name!r}>'

    def __str__(self):
        '''
        >>> table = Table('customers')
        >>> table.add_column(Column('id', 'INTEGER'))
        >>> table.add_column(Column('name', 'VARCHAR2'))
        >>> print(table)
        public.customers(id, name)
        '''

        return f'{self.schema}.{self.name}({", ".join(c.name for c in self.columns)})'

    @property
    def sql(self):
        '''
        Returns full SQL for table definition:

        CREATE TABLE "countries" (
          "code" int PRIMARY KEY,
          "name" varchar,
          "continent_name" varchar
        );

        Also returns indexes if they were defined:

        CREATE INDEX ON "products" ("id", "name");
        '''
        self.check_attributes_for_sql()
        name = self._get_full_name_for_sql()
        components = [f'CREATE TABLE {name} (']

        body = []
        body.extend(indent(c.sql, 2) for c in self.columns)
        body.extend(indent(i.sql, 2) for i in self.indexes if i.pk)
        body.extend(indent(r.sql, 2) for r in self._get_references_for_sql())

        if self._has_composite_pk():
            body.append(
                "  PRIMARY KEY ("
                + ', '.join(f'"{c.name}"' for c in self.columns if c.pk)
                + ')')
        components.append(',\n'.join(body))
        components.append(');')
        components.extend('\n' + i.sql for i in self.indexes if not i.pk)

        result = comment_to_sql(self.comment) if self.comment else ''
        result += '\n'.join(components)

        if self.note:
            result += f'\n\n{self.note.sql}'

        for col in self.columns:
            if col.note:
                quoted_note = f"'{col.note.text}'"
                note_sql = f'COMMENT ON COLUMN "{self.name}"."{col.name}" IS {quoted_note};'
                result += f'\n\n{note_sql}'
        return result

    @property
    def dbml(self):
        result = comment_to_dbml(self.comment) if self.comment else ''

        name = self._get_full_name_for_sql()

        result += f'Table {name} '
        if self.alias:
            result += f'as "{self.alias}" '
        result += '{\n'
        columns_str = '\n'.join(c.dbml for c in self.columns)
        result += indent(columns_str) + '\n'
        if self.note:
            result += indent(self.note.dbml) + '\n'
        if self.indexes:
            result += '\n    indexes {\n'
            indexes_str = '\n'.join(i.dbml for i in self.indexes)
            result += indent(indexes_str, 8) + '\n'
            result += '    }\n'

        result += '}'
        return result



from .column import Column
from .enum import Enum
from .enum import EnumItem
from .expression import Expression
from .index import Index
from .note import Note
from .project import Project
from .reference import Reference
from .table import Table
from .table_group import TableGroup



from typing import List
from typing import Optional

from .table import Table
from pydbml.tools import comment_to_dbml


class TableGroup:
    '''
    TableGroup `items` parameter initially holds just the names of the tables,
    but after parsing the whole document, PyDBMLParseResults class replaces
    them with references to actual tables.
    '''

    def __init__(self,
                 name: str,
                 items: List[Table],
                 comment: Optional[str] = None):
        self.database = None
        self.name = name
        self.items = items
        self.comment = comment

    def __repr__(self):
        """
        >>> tg = TableGroup('mygroup', ['t1', 't2'])
        >>> tg
        <TableGroup 'mygroup', ['t1', 't2']>
        >>> t1 = Table('t1')
        >>> t2 = Table('t2')
        >>> tg.items = [t1, t2]
        >>> tg
        <TableGroup 'mygroup', ['t1', 't2']>
        """

        items = [i if isinstance(i, str) else i.name for i in self.items]
        return f'<TableGroup {self.name!r}, {items!r}>'

    def __getitem__(self, key: int) -> Table:
        return self.items[key]

    def __iter__(self):
        return iter(self.items)

    @property
    def dbml(self):

        result = comment_to_dbml(self.comment) if self.comment else ''
        result += f'TableGroup {self.name} {{\n'
        for i in self.items:
            result += f'    {i._get_full_name_for_sql()}\n'
        result += '}'
        return result



from typing import Dict
from typing import Optional
from typing import Union

from .note import Note
from pydbml.tools import comment_to_dbml
from pydbml.tools import indent


class Project:
    def __init__(self,
                 name: str,
                 items: Optional[Dict[str, str]] = None,
                 note: Optional[Union['Note', str]] = None,
                 comment: Optional[str] = None):
        self.database = None
        self.name = name
        self.items = items
        self.note = Note(note)
        self.comment = comment

    def __repr__(self):
        """
        >>> Project('myproject')
        <Project 'myproject'>
        """

        return f'<Project {self.name!r}>'

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    @property
    def dbml(self):
        result = comment_to_dbml(self.comment) if self.comment else ''
        result += f'Project "{self.name}" {{\n'
        if self.items:
            items_str = ''
            for k, v in self.items.items():
                if '\n' in v:
                    items_str += f"{k}: '''{v}'''\n"
                else:
                    items_str += f"{k}: '{v}'\n"
            result += indent(items_str[:-1]) + '\n'
        if self.note:
            result += indent(self.note.dbml) + '\n'
        result += '}'
        return result



from typing import Any
from typing import Tuple

from pydbml.exceptions import AttributeMissingError


class SQLObject:
    '''
    Base class for all SQL objects.
    '''
    required_attributes: Tuple[str, ...] = ()

    def check_attributes_for_sql(self):
        '''
        Check if all attributes, required for rendering SQL are set in the
        instance. If some attribute is missing, raise AttributeMissingError
        '''
        for attr in self.required_attributes:
            if getattr(self, attr) is None:
                raise AttributeMissingError(
                    f'Cannot render SQL. Missing required attribute "{attr}".'
                )

    def __setattr__(self, name: str, value: Any):
        """
        Required for type testing with MyPy.
        """
        super().__setattr__(name, value)

    def __eq__(self, other: object) -> bool:
        """
        Two instances of the same SQLObject subclass are equal if all their
        attributes are equal.
        """

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False



from . import classes
from .parser import PyDBML
from .database import Database
from pydbml.constants import MANY_TO_ONE
from pydbml.constants import ONE_TO_MANY
from pydbml.constants import ONE_TO_ONE



from typing import List
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from .base import SQLObject
from .column import Column
from .expression import Expression
from .note import Note
from pydbml.tools import comment_to_dbml
from pydbml.tools import comment_to_sql
from pydbml.tools import note_option_to_dbml

if TYPE_CHECKING:  # pragma: no cover
    from .table import Table


class Index(SQLObject):
    '''Class representing index.'''
    required_attributes = ('subjects', 'table')

    def __init__(self,
                 subjects: List[Union[str, 'Column', 'Expression']],
                 name: Optional[str] = None,
                 unique: bool = False,
                 type: Optional[Literal['hash', 'btree']] = None,
                 pk: bool = False,
                 note: Optional[Union['Note', str]] = None,
                 comment: Optional[str] = None):
        self.subjects = subjects
        self.table: Optional['Table'] = None

        self.name = name if name else None
        self.unique = unique
        self.type = type
        self.pk = pk
        self.note = Note(note)
        self.comment = comment

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    @property
    def subject_names(self):
        '''
        Returns updated list of subject names.
        '''
        return [s.name if isinstance(s, Column) else str(s) for s in self.subjects]

    def __repr__(self):
        '''
        >>> c = Column('col', 'int')
        >>> i = Index([c, '(c*2)'])
        >>> i
        <Index None, ['col', '(c*2)']>
        >>> from .table import Table
        >>> t = Table('test')
        >>> t.add_column(c)
        >>> t.add_index(i)
        >>> i
        <Index 'test', ['col', '(c*2)']>
        '''

        table_name = self.table.name if self.table else None
        return f"<Index {table_name!r}, {self.subject_names!r}>"

    def __str__(self):
        '''
        >>> c = Column('col', 'int')
        >>> i = Index([c, '(c*2)'])
        >>> print(i)
        Index([col, (c*2)])
        >>> from .table import Table
        >>> t = Table('test')
        >>> t.add_column(c)
        >>> t.add_index(i)
        >>> print(i)
        Index(test[col, (c*2)])
        '''

        table_name = self.table.name if self.table else ''
        subjects = ', '.join(self.subject_names)
        return f"Index({table_name}[{subjects}])"

    @property
    def sql(self):
        '''
        Returns inline SQL of the index to be created separately from table
        definition:

        CREATE UNIQUE INDEX ON "products" USING HASH ("id");

        But if it's a (composite) primary key index, returns an inline SQL for
        composite primary key to be used inside table definition:

        PRIMARY KEY ("id", "name")

        '''
        self.check_attributes_for_sql()
        subjects = []

        for subj in self.subjects:
            if isinstance(subj, Column):
                subjects.append(f'"{subj.name}"')
            elif isinstance(subj, Expression):
                subjects.append(subj.sql)
            else:
                subjects.append(subj)
        keys = ', '.join(subj for subj in subjects)
        if self.pk:
            result = comment_to_sql(self.comment) if self.comment else ''
            result += f'PRIMARY KEY ({keys})'
            return result

        components = ['CREATE']
        if self.unique:
            components.append('UNIQUE')
        components.append('INDEX')
        if self.name:
            components.append(f'"{self.name}"')
        components.append(f'ON "{self.table.name}"')
        if self.type:
            components.append(f'USING {self.type.upper()}')
        components.append(f'({keys})')
        result = comment_to_sql(self.comment) if self.comment else ''
        result += ' '.join(components) + ';'
        return result

    @property
    def dbml(self):
        subjects = []

        for subj in self.subjects:
            if isinstance(subj, Column):
                subjects.append(subj.name)
            elif isinstance(subj, Expression):
                subjects.append(subj.dbml)
            else:
                subjects.append(subj)

        result = comment_to_dbml(self.comment) if self.comment else ''

        if len(subjects) > 1:
            result += f'({", ".join(subj for subj in subjects)})'
        else:
            result += subjects[0]

        options = []
        if self.name:
            options.append(f"name: '{self.name}'")
        if self.pk:
            options.append('pk')
        if self.unique:
            options.append('unique')
        if self.type:
            options.append(f'type: {self.type}')
        if self.note:
            options.append(note_option_to_dbml(self.note))

        if options:
            result += f' [{", ".join(options)}]'
        return result



import re
from typing import Any

from .base import SQLObject
from pydbml.tools import indent
from pydbml import classes


class Note(SQLObject):

    def __init__(self, text: Any):
        self.text: str
        if isinstance(text, Note):
            self.text = text.text
        else:
            self.text = str(text) if text else ''
        self.parent: Any = None

    def __str__(self):
        '''
        >>> print(Note('Note text'))
        Note text
        '''

        return self.text

    def __bool__(self):
        return bool(self.text)

    def __repr__(self):
        '''
        >>> Note('Note text')
        Note('Note text')
        '''

        return f'Note({repr(self.text)})'

    def _prepare_text_for_sql(self) -> str:
        '''
        - Process special escape sequence: slash before line break, which means no line break
        https://www.dbml.org/docs/#multi-line-string

        - replace all single quotes with double quotes
        '''

        pattern = re.compile(r'\\\n')
        result = pattern.sub('', self.text)

        result = result.replace("'", '"')
        return result

    def _prepare_text_for_dbml(self):
        '''Escape single quotes'''
        pattern = re.compile(r"('''|')")
        return pattern.sub(r'\\\1', self.text)

    def generate_comment_on(self, entity: str, name: str) -> str:
        """Generate a COMMENT ON clause out from this note."""
        quoted_text = f"'{self._prepare_text_for_sql()}'"
        note_sql = f'COMMENT ON {entity.upper()} "{name}" IS {quoted_text};'
        return note_sql

    @property
    def sql(self):
        """
        For Tables and Columns Note is converted into COMMENT ON clause. All other entities don't
        have notes generated in their SQL code, but as a fallback their notes are rendered as SQL
        comments when sql property is called directly.
        """
        if self.text:
            if isinstance(self.parent, (classes.Table, classes.Column)):
                return self.generate_comment_on(self.parent.__class__.__name__, self.parent.name)
            else:
                text = self._prepare_text_for_sql()
                return '\n'.join(f'-- {line}' for line in text.split('\n'))
        else:
            return ''

    @property
    def dbml(self):
        text = self._prepare_text_for_dbml()
        if '\n' in text:
            note_text = f"'''\n{text}\n'''"
        else:
            note_text = f"'{text}'"

        note_text = indent(note_text)
        result = f'Note {{\n{note_text}\n}}'
        return result



import doctest

from pydbml import database
from pydbml.classes import column
from pydbml.classes import enum
from pydbml.classes import expression
from pydbml.classes import index
from pydbml.classes import note
from pydbml.classes import project
from pydbml.classes import reference
from pydbml.classes import table
from pydbml.classes import table_group
from pydbml.parser import parser


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(column))
    tests.addTests(doctest.DocTestSuite(enum))
    tests.addTests(doctest.DocTestSuite(expression))
    tests.addTests(doctest.DocTestSuite(index))
    tests.addTests(doctest.DocTestSuite(project))
    tests.addTests(doctest.DocTestSuite(note))
    tests.addTests(doctest.DocTestSuite(reference))
    tests.addTests(doctest.DocTestSuite(database))
    tests.addTests(doctest.DocTestSuite(table))
    tests.addTests(doctest.DocTestSuite(table_group))
    tests.addTests(doctest.DocTestSuite(parser))
    return tests



from itertools import chain
from typing import Collection
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from pydbml.constants import MANY_TO_MANY
from pydbml.constants import MANY_TO_ONE
from pydbml.constants import ONE_TO_MANY
from pydbml.constants import ONE_TO_ONE
from pydbml.exceptions import DBMLError
from pydbml.exceptions import TableNotFoundError
from pydbml.tools import comment_to_dbml
from pydbml.tools import comment_to_sql
from .base import SQLObject
from .column import Column
from .table import Table


class Reference(SQLObject):
    '''
    Class, representing a foreign key constraint.
    It is a separate object, which is not connected to Table or Column objects
    and its `sql` property contains the ALTER TABLE clause.
    '''
    required_attributes = ('type', 'col1', 'col2')

    def __init__(self,
                 type: Literal['>', '<', '-', '<>'],
                 col1: Union[Column, Collection[Column]],
                 col2: Union[Column, Collection[Column]],
                 name: Optional[str] = None,
                 comment: Optional[str] = None,
                 on_update: Optional[str] = None,
                 on_delete: Optional[str] = None,
                 inline: bool = False):
        self.database = None
        self.type = type
        self.col1 = [col1] if isinstance(col1, Column) else list(col1)
        self.col2 = [col2] if isinstance(col2, Column) else list(col2)
        self.name = name if name else None
        self.comment = comment
        self.on_update = on_update
        self.on_delete = on_delete
        self.inline = inline

    @property
    def join_table(self) -> Optional['Table']:
        if self.type != MANY_TO_MANY:
            return None

        if self.table1 is None:
            raise TableNotFoundError(f"Cannot generate join table for {self}: table 1 is unknown")
        if self.table2 is None:
            raise TableNotFoundError(f"Cannot generate join table for {self}: table 2 is unknown")

        return Table(
            name=f'{self.table1.name}_{self.table2.name}',
            columns=(
                Column(name=f'{c.table.name}_{c.name}', type=c.type, not_null=True, pk=True)  # type: ignore
                for c in chain(self.col1, self.col2)
            ),
            abstract=True
        )

    @property
    def table1(self) -> Optional['Table']:
        self._validate()
        return self.col1[0].table if self.col1 else None

    @property
    def table2(self) -> Optional['Table']:
        self._validate()
        return self.col2[0].table if self.col2 else None

    def __repr__(self):
        '''
        >>> c1 = Column('c1', 'int')
        >>> c2 = Column('c2', 'int')
        >>> Reference('>', col1=c1, col2=c2)
        <Reference '>', ['c1'], ['c2']>
        >>> c12 = Column('c12', 'int')
        >>> c22 = Column('c22', 'int')
        >>> Reference('<', col1=[c1, c12], col2=(c2, c22))
        <Reference '<', ['c1', 'c12'], ['c2', 'c22']>
        '''

        col1 = ', '.join(f'{c.name!r}' for c in self.col1)
        col2 = ', '.join(f'{c.name!r}' for c in self.col2)
        return f"<Reference {self.type!r}, [{col1}], [{col2}]>"

    def __str__(self):
        '''
        >>> c1 = Column('c1', 'int')
        >>> c2 = Column('c2', 'int')
        >>> print(Reference('>', col1=c1, col2=c2))
        Reference([c1] > [c2]
        >>> c12 = Column('c12', 'int')
        >>> c22 = Column('c22', 'int')
        >>> print(Reference('<', col1=[c1, c12], col2=(c2, c22)))
        Reference([c1, c12] < [c2, c22]
        '''

        col1 = ', '.join(f'{c.name}' for c in self.col1)
        col2 = ', '.join(f'{c.name}' for c in self.col2)
        return f"Reference([{col1}] {self.type} [{col2}]"

    def _validate(self):
        table1 = self.col1[0].table
        if any(c.table != table1 for c in self.col1):
            raise DBMLError('Columns in col1 are from different tables')

        table2 = self.col2[0].table
        if any(c.table != table2 for c in self.col2):
            raise DBMLError('Columns in col2 are from different tables')

    def _validate_for_sql(self):
        for col in chain(self.col1, self.col2):
            if col.table is None:
                raise TableNotFoundError(f'Table on {col} is not set')

    def _generate_inline_sql(self, source_col: List['Column'], ref_col: List['Column']) -> str:
        result = comment_to_sql(self.comment) if self.comment else ''
        result += (
            f'{{c}}FOREIGN KEY ({self._col_names(source_col)}) '  # type: ignore
            f'REFERENCES {ref_col[0].table._get_full_name_for_sql()} ({self._col_names(ref_col)})'
        )
        if self.on_update:
            result += f' ON UPDATE {self.on_update.upper()}'
        if self.on_delete:
            result += f' ON DELETE {self.on_delete.upper()}'
        return result

    def _generate_not_inline_sql(self, c1: List['Column'], c2: List['Column']):
        result = comment_to_sql(self.comment) if self.comment else ''
        result += (
            f'ALTER TABLE {c1[0].table._get_full_name_for_sql()}'  # type: ignore
            f' ADD {{c}}FOREIGN KEY ({self._col_names(c1)})'
            f' REFERENCES {c2[0].table._get_full_name_for_sql()} ({self._col_names(c2)})'
        )
        if self.on_update:
            result += f' ON UPDATE {self.on_update.upper()}'
        if self.on_delete:
            result += f' ON DELETE {self.on_delete.upper()}'
        return result + ';'

    def _generate_many_to_many_sql(self) -> str:
        join_table = self.join_table
        table_sql = join_table.sql  # type: ignore

        n = len(self.col1)
        ref1_sql = self._generate_not_inline_sql(join_table.columns[:n], self.col1)  # type: ignore
        ref2_sql = self._generate_not_inline_sql(join_table.columns[n:], self.col2)  # type: ignore

        result = '\n\n'.join((table_sql, ref1_sql, ref2_sql))
        return result.format(c='')

    @staticmethod
    def _col_names(cols: List[Column]) -> str:
        return ', '.join(f'"{c.name}"' for c in cols)

    @property
    def sql(self) -> str:
        '''
        Returns SQL of the reference:

        ALTER TABLE "orders" ADD FOREIGN KEY ("customer_id") REFERENCES "customers ("id");

        '''
        self.check_attributes_for_sql()
        self._validate_for_sql()

        if self.type == MANY_TO_MANY:
            return self._generate_many_to_many_sql()

        result = ''
        if self.inline:
            if self.type in (MANY_TO_ONE, ONE_TO_ONE):
                result = self._generate_inline_sql(self.col1, self.col2)
            elif self.type == ONE_TO_MANY:
                result = self._generate_inline_sql(self.col2, self.col1)
        else:
            if self.type in (MANY_TO_ONE, ONE_TO_ONE):
                result = self._generate_not_inline_sql(c1=self.col1, c2=self.col2)
            elif self.type == ONE_TO_MANY:
                result = self._generate_not_inline_sql(c1=self.col2, c2=self.col1)

        c = f'CONSTRAINT "{self.name}" ' if self.name else ''

        return result.format(c=c)

    @property
    def dbml(self) -> str:
        self._validate_for_sql()
        if self.inline:
            # settings are ignored for inline ref
            if len(self.col2) > 1:
                raise DBMLError('Cannot render DBML: composite ref cannot be inline')
            table_name = self.col2[0].table._get_full_name_for_sql()  # type: ignore
            return f'ref: {self.type} {table_name}."{self.col2[0].name}"'
        else:
            result = comment_to_dbml(self.comment) if self.comment else ''
            result += 'Ref'
            if self.name:
                result += f' {self.name}'

            if len(self.col1) == 1:
                col1 = f'"{self.col1[0].name}"'
            else:
                names = (f'"{c.name}"' for c in self.col1)
                col1 = f'({", ".join(names)})'

            if len(self.col2) == 1:
                col2 = f'"{self.col2[0].name}"'
            else:
                names = (f'"{c.name}"' for c in self.col2)
                col2 = f'({", ".join(names)})'

            options = []
            if self.on_update:
                options.append(f'update: {self.on_update}')
            if self.on_delete:
                options.append(f'delete: {self.on_delete}')

            options_str = f' [{", ".join(options)}]' if options else ''
            result += (
                ' {\n    '  # type: ignore
                f'{self.table1._get_full_name_for_sql()}.{col1} '
                f'{self.type} '
                f'{self.table2._get_full_name_for_sql()}.{col2}'
                f'{options_str}'
                '\n}'
            )
            return result



import pyparsing as pp

from .common import _
from .common import _c
from .common import end
from .generic import name
from pydbml.parser.blueprints import TableGroupBlueprint

pp.ParserElement.set_default_whitespace_chars(' \t\r')

table_name = pp.Combine(name + '.' + name) | name

table_group = _c + (
    pp.CaselessLiteral('TableGroup')
    - name('name') + _
    - '{' + _
    - (table_name + _)[...]('items') + _
    - '}'
) + end


def parse_table_group(s, loc, tok):
    '''
    TableGroup tablegroup_name {
        table1
        table2
        table3
    }
    '''
    init_dict = {
        'name': tok['name'],
        'items': list(tok.get('items', []))
    }
    if 'comment_before' in tok:
        comment = '\n'.join(c[0] for c in tok['comment_before'])
        init_dict['comment'] = comment
    return TableGroupBlueprint(**init_dict)


table_group.set_parse_action(parse_table_group)



import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .classes import Note


def comment(val: str, comb: str) -> str:
    return '\n'.join(f'{comb} {cl}' for cl in val.split('\n')) + '\n'


def comment_to_dbml(val: str) -> str:
    return comment(val, '//')


def comment_to_sql(val: str) -> str:
    return comment(val, '--')


def note_option_to_dbml(val: 'Note') -> str:
    if '\n' in val.text:
        return f"note: '''{val.text}'''"
    else:
        return f"note: '{val.text}'"


def indent(val: str, spaces=4) -> str:
    if val == '':
        return val
    return ' ' * spaces + val.replace('\n', '\n' + ' ' * spaces)


def remove_bom(source: str) -> str:
    if source and source[0] == '\ufeff':
        source = source[1:]
    return source


def strip_empty_lines(source: str) -> str:
    """Remove empty lines or lines with just spaces from beginning and end."""
    first_line = 0
    lines = source.split('\n')
    last_line = len(lines) - 1
    while not lines[first_line] or lines[first_line].isspace():
        first_line += 1
    while not lines[last_line] or lines[last_line].isspace():
        last_line -= 1
    return '\n'.join(lines[first_line: last_line + 1])


def remove_indentation(source: str) -> str:
    pattern = re.compile(r'^\s*')

    lines = source.split('\n')
    spaces = []
    for line in lines:
        if line and not line.isspace():
            indent_match = pattern.search(line)
            if indent_match is not None:  # this is just for you mypy
                spaces.append(len(indent_match[0]))

    indent = min(spaces)
    lines = [l[indent:] for l in lines]
    return '\n'.join(lines)



from unittest import TestCase

from pydbml.classes import Note
from pydbml.tools import comment_to_dbml
from pydbml.tools import comment_to_sql
from pydbml.tools import indent
from pydbml.tools import note_option_to_dbml


class TestCommentToDBML(TestCase):
    def test_comment(self) -> None:
        oneline = 'comment'
        self.assertEqual(f'// {oneline}\n', comment_to_dbml(oneline))

        expected = \
'''// 
// line1
// line2
// line3
// 
'''
        source = '\nline1\nline2\nline3\n'
        self.assertEqual(comment_to_dbml(source), expected)


class TestCommentToSQL(TestCase):
    def test_comment(self) -> None:
        oneline = 'comment'
        self.assertEqual(f'-- {oneline}\n', comment_to_sql(oneline))

        expected = \
'''-- 
-- line1
-- line2
-- line3
-- 
'''
        source = '\nline1\nline2\nline3\n'
        self.assertEqual(comment_to_sql(source), expected)


class TestNoteOptionToDBML(TestCase):
    def test_oneline(self) -> None:
        note = Note('one line note')
        self.assertEqual(f"note: 'one line note'", note_option_to_dbml(note))

    def test_multiline(self) -> None:
        note = Note('line1\nline2\nline3')
        expected = "note: '''line1\nline2\nline3'''"
        self.assertEqual(expected, note_option_to_dbml(note))


class TestIndent(TestCase):
    def test_empty(self) -> None:
        self.assertEqual(indent(''), '')

    def test_nonempty(self) -> None:
        oneline = 'one line text'
        self.assertEqual(indent(oneline), f'    {oneline}')
        source = 'line1\nline2\nline3'
        expected = '    line1\n    line2\n    line3'
        self.assertEqual(indent(source), expected)
        expected2 = '  line1\n  line2\n  line3'
        self.assertEqual(indent(source, 2), expected2)



from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from .base import SQLObject
from .note import Note
from pydbml.tools import comment_to_dbml
from pydbml.tools import comment_to_sql
from pydbml.tools import indent
from pydbml.tools import note_option_to_dbml


class EnumItem:
    '''Single enum item'''

    def __init__(self,
                 name: str,
                 note: Optional[Union['Note', str]] = None,
                 comment: Optional[str] = None):
        self.name = name
        self.note = Note(note)
        self.comment = comment

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, val: Note) -> None:
        self._note = val
        val.parent = self

    def __repr__(self):
        '''
        >>> EnumItem('en-US')
        <EnumItem 'en-US'>
        '''

        return f'<EnumItem {self.name!r}>'

    def __str__(self):
        '''
        >>> print(EnumItem('en-US'))
        en-US
        '''

        return self.name

    @property
    def sql(self):
        result = comment_to_sql(self.comment) if self.comment else ''
        result += f"'{self.name}',"
        return result

    @property
    def dbml(self):
        result = comment_to_dbml(self.comment) if self.comment else ''
        result += f'"{self.name}"'
        if self.note:
            result += f' [{note_option_to_dbml(self.note)}]'
        return result


class Enum(SQLObject):
    required_attributes = ('name', 'schema', 'items')

    def __init__(self,
                 name: str,
                 items: Iterable[Union['EnumItem', str]],
                 schema: str = 'public',
                 comment: Optional[str] = None):
        self.database = None
        self.name = name
        self.schema = schema
        self.comment = comment
        self.items: List[EnumItem] = []
        for item in items:
            self.add_item(item)

    def add_item(self, item: Union['EnumItem', str]) -> None:
        if isinstance(item, EnumItem):
            self.items.append(item)
        elif isinstance(item, str):
            self.items.append(EnumItem(item))

    def __getitem__(self, key: int) -> EnumItem:
        return self.items[key]

    def __iter__(self):
        return iter(self.items)

    def __repr__(self):
        '''
        >>> en = EnumItem('en-US')
        >>> ru = EnumItem('ru-RU')
        >>> Enum('languages', [en, ru])
        <Enum 'languages', ['en-US', 'ru-RU']>
        '''

        item_names = [i.name for i in self.items]
        classname = self.__class__.__name__
        return f'<{classname} {self.name!r}, {item_names!r}>'

    def __str__(self):
        '''
        >>> en = EnumItem('en-US')
        >>> ru = EnumItem('ru-RU')
        >>> print(Enum('languages', [en, ru]))
        languages
        '''

        return self.name

    def _get_full_name_for_sql(self) -> str:
        if self.schema == 'public':
            return f'"{self.name}"'
        else:
            return f'"{self.schema}"."{self.name}"'

    @property
    def sql(self):
        '''
        Returns SQL for enum type:

        CREATE TYPE "job_status" AS ENUM (
          'created',
          'running',
          'donef',
          'failure',
        );

        '''
        self.check_attributes_for_sql()
        result = comment_to_sql(self.comment) if self.comment else ''
        result += f'CREATE TYPE {self._get_full_name_for_sql()} AS ENUM (\n'
        result += '\n'.join(f'{indent(i.sql, 2)}' for i in self.items)
        result += '\n);'
        return result

    @property
    def dbml(self):
        result = comment_to_dbml(self.comment) if self.comment else ''
        result += f'Enum {self._get_full_name_for_sql()} {{\n'
        items_str = '\n'.join(i.dbml for i in self.items)
        result += indent(items_str)
        result += '\n}'
        return result
